# pytest --maxfail=1 --disable-warnings --no-cov -s -vv acapy_agent/database_manager/wql_normalized/tests/test_sqlite_TagsqlEncoder_All_key_value.py
# python -m unittest acapy_agent/database_manager/wql_normalized/tests/test_sqlite_TagsqlEncoder_All_key_value.py -v
import logging
import sqlite3
import unittest

from acapy_agent.database_manager.wql_normalized.encoders import encoder_factory
from acapy_agent.database_manager.wql_normalized.query import query_from_str
from acapy_agent.database_manager.wql_normalized.tags import (
    TagName,
    TagQuery,
    query_to_tagquery,
)

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def replace_placeholders(query, args):
    """Replace each '?' in the query with the corresponding argument.

    Properly quote arguments for SQL, escaping single quotes by doubling them.
    Example: 'O'Reilly' becomes 'O''Reilly'.
    """
    parts = query.split("?")
    if len(parts) - 1 != len(args):
        raise ValueError("Number of placeholders does not match number of arguments")
    result = parts[0]
    for part, arg in zip(parts[1:], args):
        escaped_arg = str(arg).replace("'", "''")  # Escape single quotes for SQL
        result += f"'{escaped_arg}'" + part
    return result


class TestSqliteTagEncoderNonNormalized(unittest.TestCase):
    """Test cases for the SqliteTagEncoder class in non-normalized mode."""

    def setUp(self):
        """Set up SQLite in-memory database and encoder."""
        self.enc_name = lambda x: x  # Return tag names as strings
        self.enc_value = lambda x: x  # Return tag values as strings
        try:
            self.conn = sqlite3.connect(":memory:")
            self.cursor = self.conn.cursor()
            # Create tables for key-value pair structure
            self.cursor.execute("""
                CREATE TABLE items (
                    id INTEGER PRIMARY KEY
                )
            """)
            self.cursor.execute("""
                CREATE TABLE items_tags (
                    item_id INTEGER,
                    name TEXT,
                    value TEXT,
                    FOREIGN KEY(item_id) REFERENCES items(id)
                )
            """)
            self.conn.commit()
            logger.info("Tables 'items' and 'items_tags' created in setUp")
            self.encoder = encoder_factory.get_encoder(
                "sqlite", self.enc_name, self.enc_value, normalized=False
            )
        except Exception as e:
            logger.error(f"Failed to set up SQLite database: {e}")
            raise

    def tearDown(self):
        """Clean up by closing the SQLite connection."""
        try:
            self.conn.close()
            logger.info("SQLite connection closed in tearDown")
        except Exception as e:
            logger.error(f"Failed to tear down SQLite connection: {e}")
            raise

    def run_query_and_verify(self, sql_query, params, expected_ids, test_name):
        try:
            self.cursor.execute(f"SELECT i.id FROM items i WHERE {sql_query}", params)
            actual_ids = sorted([row[0] for row in self.cursor.fetchall()])
            self.assertEqual(
                actual_ids,
                expected_ids,
                f"{test_name} failed: Expected IDs {expected_ids}, got {actual_ids}",
            )
        except Exception as e:
            logger.error(f"Query execution failed in {test_name}: {e}")
            raise

    def verify_round_trip(self, query, original_sql_query, original_params):
        """Verify that converting TagQuery to WQL and back results in the same SQLite query."""
        wql_str = query.to_wql_str()
        parsed_query = query_from_str(wql_str)
        parsed_tag_query = query_to_tagquery(parsed_query)
        parsed_sql_query = self.encoder.encode_query(parsed_tag_query)
        parsed_params = self.encoder.arguments
        self.assertEqual(
            (original_sql_query, original_params),
            (parsed_sql_query, parsed_params),
            f"Round-trip SQLite query mismatch in {self._testMethodName}",
        )

    # Individual Operator Tests
    def test_eq_positive(self):
        query = TagQuery.eq(TagName("field"), "value")
        wql = query.to_wql_str()
        print(f"Test: Positive equality query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?)"
        )
        expected_params = ["field", "value"]
        self.assertEqual(sql_query, expected_query, "Positive equality query mismatch")
        self.assertEqual(params, expected_params, "Positive equality params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany("INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,)])
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "value"), (2, "field", "other"), (3, "field", "value")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'other'), "
            "(3, 'field', 'value');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 3")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 3], "Positive equality")

    def test_eq_negated(self):
        query = TagQuery.not_(TagQuery.eq(TagName("field"), "value"))
        wql = query.to_wql_str()
        print(f"Test: Negated equality query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?)"
        )
        expected_params = ["field", "value"]
        self.assertEqual(sql_query, expected_query, "Negated equality query mismatch")
        self.assertEqual(params, expected_params, "Negated equality params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany("INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,)])
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "value"), (2, "field", "other"), (3, "field", "value")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'other'), "
            "(3, 'field', 'value');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 2")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [2], "Negated equality")

    def test_neq_positive(self):
        query = TagQuery.neq(TagName("field"), "value")
        wql = query.to_wql_str()
        print(f"Test: Positive inequality query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value != ?)"
        )
        expected_params = ["field", "value"]
        self.assertEqual(sql_query, expected_query, "Positive inequality query mismatch")
        self.assertEqual(params, expected_params, "Positive inequality params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany("INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,)])
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "value"), (2, "field", "other"), (3, "field", "different")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'other'), "
            "(3, 'field', 'different');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 2, 3")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [2, 3], "Positive inequality")

    def test_neq_negated(self):
        query = TagQuery.not_(TagQuery.neq(TagName("field"), "value"))
        wql = query.to_wql_str()
        print(f"Test: Negated inequality query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value != ?)"
        )
        expected_params = ["field", "value"]
        self.assertEqual(sql_query, expected_query, "Negated inequality query mismatch")
        self.assertEqual(params, expected_params, "Negated inequality params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany("INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,)])
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "value"), (2, "field", "other"), (3, "field", "value")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'other'), "
            "(3, 'field', 'value');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 3")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 3], "Negated inequality")

    def test_gt_positive(self):
        query = TagQuery.gt(TagName("price"), "100")
        wql = query.to_wql_str()
        print(f"Test: Positive greater-than query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value > ?)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Positive greater-than query mismatch"
        )
        self.assertEqual(params, expected_params, "Positive greater-than params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 3, 4")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [3, 4], "Positive greater-than")

    def test_gt_negated(self):
        query = TagQuery.not_(TagQuery.gt(TagName("price"), "100"))
        wql = query.to_wql_str()
        print(f"Test: Negated greater-than query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value > ?)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(sql_query, expected_query, "Negated greater-than query mismatch")
        self.assertEqual(params, expected_params, "Negated greater-than params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 2], "Negated greater-than")

    def test_gte_positive(self):
        query = TagQuery.gte(TagName("price"), "100")
        wql = query.to_wql_str()
        print(f"Test: Positive greater-than-or-equal query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value >= ?)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Positive greater-than-or-equal query mismatch"
        )
        self.assertEqual(
            params, expected_params, "Positive greater-than-or-equal params mismatch"
        )
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 2, 3, 4")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(
            sql_query, params, [2, 3, 4], "Positive greater-than-or-equal"
        )

    def test_gte_negated(self):
        query = TagQuery.not_(TagQuery.gte(TagName("price"), "100"))
        wql = query.to_wql_str()
        print(f"Test: Negated greater-than-or-equal query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value >= ?)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Negated greater-than-or-equal query mismatch"
        )
        self.assertEqual(
            params, expected_params, "Negated greater-than-or-equal params mismatch"
        )
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 1")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1], "Negated greater-than-or-equal")

    def test_lt_positive(self):
        query = TagQuery.lt(TagName("price"), "100")
        wql = query.to_wql_str()
        print(f"Test: Positive less-than query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value < ?)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(sql_query, expected_query, "Positive less-than query mismatch")
        self.assertEqual(params, expected_params, "Positive less-than params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 1")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1], "Positive less-than")

    def test_lt_negated(self):
        query = TagQuery.not_(TagQuery.lt(TagName("price"), "100"))
        wql = query.to_wql_str()
        print(f"Test: Negated less-than query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value < ?)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(sql_query, expected_query, "Negated less-than query mismatch")
        self.assertEqual(params, expected_params, "Negated less-than params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 2, 3, 4")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [2, 3, 4], "Negated less-than")

    def test_lte_positive(self):
        query = TagQuery.lte(TagName("price"), "100")
        wql = query.to_wql_str()
        print(f"Test: Positive less-than-or-equal query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value <= ?)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Positive less-than-or-equal query mismatch"
        )
        self.assertEqual(
            params, expected_params, "Positive less-than-or-equal params mismatch"
        )
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(
            sql_query, params, [1, 2], "Positive less-than-or-equal"
        )

    def test_lte_negated(self):
        query = TagQuery.not_(TagQuery.lte(TagName("price"), "100"))
        wql = query.to_wql_str()
        print(f"Test: Negated less-than-or-equal query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value <= ?)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Negated less-than-or-equal query mismatch"
        )
        self.assertEqual(
            params, expected_params, "Negated less-than-or-equal params mismatch"
        )
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 3, 4")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [3, 4], "Negated less-than-or-equal")

    def test_like_positive(self):
        query = TagQuery.like(TagName("field"), "%pat%")
        wql = query.to_wql_str()
        print(f"Test: Positive LIKE query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value LIKE ?)"
        )
        expected_params = ["field", "%pat%"]
        self.assertEqual(sql_query, expected_query, "Positive LIKE query mismatch")
        self.assertEqual(params, expected_params, "Positive LIKE params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "field", "pattern"),
                (2, "field", "path"),
                (3, "field", "other"),
                (4, "field", "pat"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'pattern'), "
            "(2, 'field', 'path'), "
            "(3, 'field', 'other'), "
            "(4, 'field', 'pat');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2, 4")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 2, 4], "Positive LIKE")

    def test_like_negated(self):
        query = TagQuery.not_(TagQuery.like(TagName("field"), "%pat%"))
        wql = query.to_wql_str()
        print(f"Test: Negated LIKE query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value LIKE ?)"
        )
        expected_params = ["field", "%pat%"]
        self.assertEqual(sql_query, expected_query, "Negated LIKE query mismatch")
        self.assertEqual(params, expected_params, "Negated LIKE params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "field", "pattern"),
                (2, "field", "path"),
                (3, "field", "other"),
                (4, "field", "pat"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'pattern'), "
            "(2, 'field', 'path'), "
            "(3, 'field', 'other'), "
            "(4, 'field', 'pat');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 3")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [3], "Negated LIKE")

    def test_in_positive(self):
        query = TagQuery.in_(TagName("field"), ["a", "b"])
        wql = query.to_wql_str()
        print(f"Test: Positive IN query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value IN (?, ?))"
        )
        expected_params = ["field", "a", "b"]
        self.assertEqual(sql_query, expected_query, "Positive IN query mismatch")
        self.assertEqual(params, expected_params, "Positive IN params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "a"), (2, "field", "b"), (3, "field", "c"), (4, "field", "a")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'a'), "
            "(2, 'field', 'b'), "
            "(3, 'field', 'c'), "
            "(4, 'field', 'a');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2, 4")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 2, 4], "Positive IN")

    def test_in_negated(self):
        query = TagQuery.not_(TagQuery.in_(TagName("field"), ["a", "b"]))
        wql = query.to_wql_str()
        print(f"Test: Negated IN query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value NOT IN (?, ?))"
        expected_params = ["field", "a", "b"]
        self.assertEqual(sql_query, expected_query, "Negated IN query mismatch")
        self.assertEqual(params, expected_params, "Negated IN params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "a"), (2, "field", "b"), (3, "field", "c"), (4, "field", "d")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'a'), "
            "(2, 'field', 'b'), "
            "(3, 'field', 'c'), "
            "(4, 'field', 'd');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 3, 4")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [3, 4], "Negated IN")

    def test_exist_positive(self):
        query = TagQuery.exist([TagName("field")])
        wql = query.to_wql_str()
        print(f"Test: Positive EXIST query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "i.id IN (SELECT item_id FROM items_tags WHERE name = ?)"
        expected_params = ["field"]
        self.assertEqual(sql_query, expected_query, "Positive EXIST query mismatch")
        self.assertEqual(params, expected_params, "Positive EXIST params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany("INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,)])
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "value"), (3, "field", "another")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(3, 'field', 'another');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 3")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 3], "Positive EXIST")

    def test_exist_negated(self):
        query = TagQuery.not_(TagQuery.exist([TagName("field")]))
        wql = query.to_wql_str()
        print(f"Test: Negated EXIST query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ?)"
        expected_params = ["field"]
        self.assertEqual(sql_query, expected_query, "Negated EXIST query mismatch")
        self.assertEqual(params, expected_params, "Negated EXIST params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany("INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,)])
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "value"), (3, "field", "another")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(3, 'field', 'another');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 2")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [2], "Negated EXIST")

    def test_and_multiple(self):
        query = TagQuery.and_(
            [TagQuery.eq(TagName("f1"), "v1"), TagQuery.gt(TagName("f2"), "10")]
        )
        wql = query.to_wql_str()
        print(f"Test: AND query with multiple subqueries\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "(i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?) AND i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value > ?))"
        expected_params = ["f1", "v1", "f2", "10"]
        self.assertEqual(sql_query, expected_query, "AND multiple query mismatch")
        self.assertEqual(params, expected_params, "AND multiple params mismatch")
        self.verify_round_trip(query, sql_query, params)

        # Insert items into the items table
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )

        # Insert tags into the items_tags table
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "f1", "v1"),  # Item 1: satisfies "f1" = "v1"
                (1, "f2", "15"),  # Item 1: satisfies "f2" > "10"
                (2, "f1", "05"),  # Item 2: does not satisfy "f1" = "v1"
                (3, "f2", "15"),  # Item 3: does not have "f1" = "v1"
                (4, "f1", "v1"),  # Item 4: satisfies "f1" = "v1"
                (4, "f2", "20"),  # Item 4: satisfies "f2" > "10"
            ],
        )
        self.conn.commit()

        # Print statements for debugging
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'f1', 'v1'), (1, 'f2', '15'), (2, 'f1', '05'), (3, 'f2', '15'), "
            "(4, 'f1', 'v1'), (4, 'f2', '20');"
        )
        select_query = f"SELECT i.id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 4")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")

        # Run the query and verify results
        self.run_query_and_verify(sql_query, params, [1, 4], "AND multiple")

    def test_or_multiple(self):
        query = TagQuery.or_(
            [TagQuery.eq(TagName("f1"), "v1"), TagQuery.gt(TagName("f2"), "10")]
        )
        wql = query.to_wql_str()
        print(f"Test: OR query with multiple subqueries\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "(i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?) OR i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value > ?))"
        expected_params = ["f1", "v1", "f2", "10"]
        self.assertEqual(sql_query, expected_query, "OR multiple query mismatch")
        self.assertEqual(params, expected_params, "OR multiple params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "f1", "v1"),
                (1, "f2", "15"),
                (2, "f1", "v1"),
                (2, "f2", "05"),
                (3, "f1", "v2"),
                (3, "f2", "15"),
                (4, "f1", "v2"),
                (4, "f2", "05"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'f1', 'v1'), (1, 'f2', '15'), "
            "(2, 'f1', 'v1'), (2, 'f2', '05'), "
            "(3, 'f1', 'v2'), (3, 'f2', '15'), "
            "(4, 'f1', 'v2'), (4, 'f2', '05');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2, 3")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 2, 3], "OR multiple")

    def test_nested_and_or(self):
        query = TagQuery.and_(
            [
                TagQuery.eq(TagName("f1"), "v1"),
                TagQuery.or_(
                    [TagQuery.gt(TagName("f2"), "10"), TagQuery.lt(TagName("f3"), "5")]
                ),
            ]
        )
        wql = query.to_wql_str()
        print(f"Test: Nested AND/OR query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "(i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?) AND (i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value > ?) OR i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value < ?)))"
        expected_params = ["f1", "v1", "f2", "10", "f3", "5"]
        self.assertEqual(sql_query, expected_query, "Nested AND/OR query mismatch")
        self.assertEqual(params, expected_params, "Nested AND/OR params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "f1", "v1"),
                (1, "f2", "15"),
                (1, "f3", "3"),
                (2, "f1", "v1"),
                (2, "f2", "05"),
                (2, "f3", "4"),
                (3, "f1", "v2"),
                (3, "f2", "15"),
                (3, "f3", "3"),
                (4, "f1", "v1"),
                (4, "f2", "05"),
                (4, "f3", "6"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'f1', 'v1'), (1, 'f2', '15'), (1, 'f3', '3'), "
            "(2, 'f1', 'v1'), (2, 'f2', '05'), (2, 'f3', '4'), "
            "(3, 'f1', 'v2'), (3, 'f2', '15'), (3, 'f3', '3'), "
            "(4, 'f1', 'v1'), (4, 'f2', '05'), (4, 'f3', '6');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 2], "Nested AND/OR")

    def test_comparison_conjunction(self):
        query = TagQuery.and_(
            [
                TagQuery.eq(TagName("category"), "electronics"),
                TagQuery.gt(TagName("price"), "100"),
            ]
        )
        wql = query.to_wql_str()
        print(f"Test: Comparison conjunction query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "(i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?) AND i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value > ?))"
        expected_params = ["category", "electronics", "price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Comparison conjunction query mismatch"
        )
        self.assertEqual(
            params, expected_params, "Comparison conjunction params mismatch"
        )
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "category", "electronics"),
                (1, "price", "150"),
                (2, "category", "electronics"),
                (2, "price", "090"),
                (3, "category", "books"),
                (3, "price", "120"),
                (4, "category", "electronics"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'category', 'electronics'), (1, 'price', '150'), "
            "(2, 'category', 'electronics'), (2, 'price', '090'), "
            "(3, 'category', 'books'), (3, 'price', '120'), "
            "(4, 'category', 'electronics'), (4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 4")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 4], "Comparison conjunction")

    def test_deeply_nested_not(self):
        query = TagQuery.not_(
            TagQuery.and_(
                [
                    TagQuery.or_(
                        [
                            TagQuery.eq(TagName("category"), "electronics"),
                            TagQuery.eq(TagName("sale"), "yes"),
                        ]
                    ),
                    TagQuery.not_(TagQuery.eq(TagName("stock"), "out")),
                ]
            )
        )
        wql = query.to_wql_str()
        print(f"Test: Deeply nested NOT query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "NOT ((i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?) OR i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?)) AND i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?))"
        expected_params = ["category", "electronics", "sale", "yes", "stock", "out"]
        self.assertEqual(sql_query, expected_query, "Deeply nested NOT query mismatch")
        self.assertEqual(params, expected_params, "Deeply nested NOT params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "category", "electronics"),
                (1, "stock", "in"),
                (2, "category", "electronics"),
                (2, "stock", "out"),
                (3, "sale", "yes"),
                (3, "stock", "in"),
                (4, "sale", "yes"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'category', 'electronics'), (1, 'stock', 'in'), "
            "(2, 'category', 'electronics'), (2, 'stock', 'out'), "
            "(3, 'sale', 'yes'), (3, 'stock', 'in'), "
            "(4, 'sale', 'yes');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 2")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [2], "Deeply nested NOT")

    def test_and_or_not_complex_case(self):
        query = TagQuery.not_(
            TagQuery.and_(
                [
                    TagQuery.eq(TagName("username"), "alice"),
                    TagQuery.or_(
                        [
                            TagQuery.gt(TagName("age"), "30"),
                            TagQuery.not_(TagQuery.lte(TagName("height"), "180")),
                            TagQuery.and_(
                                [
                                    TagQuery.lt(TagName("score"), "100"),
                                    TagQuery.not_(
                                        TagQuery.gte(
                                            TagName("timestamp"), "2021-01-01T00:00:00"
                                        )
                                    ),
                                ]
                            ),
                        ]
                    ),
                    TagQuery.not_(TagQuery.like(TagName("secret_code"), "abc123")),
                    TagQuery.and_(
                        [
                            TagQuery.eq(TagName("occupation"), "developer"),
                            TagQuery.not_(TagQuery.neq(TagName("status"), "active")),
                        ]
                    ),
                ]
            )
        )
        wql = query.to_wql_str()
        print(f"Test: Complex AND/OR/NOT query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "NOT (i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?) AND (i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value > ?) OR i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value <= ?) OR (i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value < ?) AND i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value >= ?))) AND i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value LIKE ?) AND (i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?) AND i.id NOT IN (SELECT item_id FROM items_tags WHERE name = ? AND value != ?)))"
        expected_params = [
            "username",
            "alice",
            "age",
            "30",
            "height",
            "180",
            "score",
            "100",
            "timestamp",
            "2021-01-01T00:00:00",
            "secret_code",
            "abc123",
            "occupation",
            "developer",
            "status",
            "active",
        ]
        self.assertEqual(sql_query, expected_query, "Complex AND/OR/NOT query mismatch")
        self.assertEqual(params, expected_params, "Complex AND/OR/NOT params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)",
            [(1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,), (11,), (12,)],
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [
                (1, "username", "bob"),
                (1, "age", "25"),
                (1, "height", "170"),
                (1, "score", "150"),
                (1, "timestamp", "2021-02-01T00:00:00"),
                (1, "secret_code", "xyz789"),
                (1, "occupation", "engineer"),
                (1, "status", "inactive"),
                (2, "username", "alice"),
                (2, "age", "35"),
                (2, "height", "190"),
                (2, "score", "90"),
                (2, "timestamp", "2020-12-01T00:00:00"),
                (2, "secret_code", "def456"),
                (2, "occupation", "developer"),
                (2, "status", "active"),
                (3, "username", "charlie"),
                (3, "age", "28"),
                (3, "height", "175"),
                (3, "score", "120"),
                (3, "timestamp", "2021-03-01T00:00:00"),
                (3, "secret_code", "ghi789"),
                (3, "occupation", "manager"),
                (3, "status", "active"),
                (4, "username", "alice"),
                (4, "age", "32"),
                (4, "height", "185"),
                (4, "score", "95"),
                (4, "timestamp", "2020-11-01T00:00:00"),
                (4, "secret_code", "abc123"),
                (4, "occupation", "developer"),
                (4, "status", "inactive"),
                (5, "username", "eve"),
                (5, "age", "40"),
                (5, "height", "160"),
                (5, "score", "85"),
                (5, "timestamp", "2021-01-15T00:00:00"),
                (5, "secret_code", "abc123"),
                (5, "occupation", "analyst"),
                (5, "status", "active"),
                (6, "username", "frank"),
                (6, "age", "29"),
                (6, "height", "182"),
                (6, "score", "105"),
                (6, "timestamp", "2020-12-15T00:00:00"),
                (6, "secret_code", "jkl012"),
                (6, "occupation", "developer"),
                (6, "status", "active"),
                (7, "username", "alice"),
                (7, "age", "33"),
                (7, "height", "195"),
                (7, "score", "88"),
                (7, "timestamp", "2020-10-01T00:00:00"),
                (7, "secret_code", "mno345"),
                (7, "occupation", "developer"),
                (7, "status", "active"),
                (8, "username", "hank"),
                (8, "age", "27"),
                (8, "height", "165"),
                (8, "score", "110"),
                (8, "timestamp", "2021-04-01T00:00:00"),
                (8, "secret_code", "pqr678"),
                (8, "occupation", "designer"),
                (8, "status", "inactive"),
                (9, "username", "alice"),
                (9, "age", "36"),
                (9, "height", "188"),
                (9, "score", "92"),
                (9, "timestamp", "2020-09-01T00:00:00"),
                (9, "secret_code", "stu901"),
                (9, "occupation", "developer"),
                (9, "status", "active"),
                (10, "username", "jack"),
                (10, "age", "31"),
                (10, "height", "179"),
                (10, "score", "115"),
                (10, "timestamp", "2021-05-01T00:00:00"),
                (10, "secret_code", "vwx234"),
                (10, "occupation", "teacher"),
                (10, "status", "active"),
                (11, "username", "kara"),
                (11, "age", "26"),
                (11, "height", "170"),
                (11, "score", "130"),
                (11, "timestamp", "2021-06-01T00:00:00"),
                (11, "secret_code", "yza567"),
                (11, "occupation", "developer"),
                (11, "status", "inactive"),
                (12, "username", "alice"),
                (12, "age", "34"),
                (12, "height", "192"),
                (12, "score", "87"),
                (12, "timestamp", "2020-08-01T00:00:00"),
                (12, "secret_code", "bcd890"),
                (12, "occupation", "developer"),
                (12, "status", "active"),
            ],
        )
        self.conn.commit()
        expected_ids = [1, 3, 4, 5, 6, 8, 10, 11]
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print(
            "INSERT INTO items (id) VALUES (1), (2), (3), (4), (5), (6), (7), (8), (9), (10), (11), (12);"
        )
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'username', 'bob'), (1, 'age', '25'), (1, 'height', '170'), (1, 'score', '150'), (1, 'timestamp', '2021-02-01T00:00:00'), (1, 'secret_code', 'xyz789'), (1, 'occupation', 'engineer'), (1, 'status', 'inactive'), "
            "(2, 'username', 'alice'), (2, 'age', '35'), (2, 'height', '190'), (2, 'score', '90'), (2, 'timestamp', '2020-12-01T00:00:00'), (2, 'secret_code', 'def456'), (2, 'occupation', 'developer'), (2, 'status', 'active'), "
            "(3, 'username', 'charlie'), (3, 'age', '28'), (3, 'height', '175'), (3, 'score', '120'), (3, 'timestamp', '2021-03-01T00:00:00'), (3, 'secret_code', 'ghi789'), (3, 'occupation', 'manager'), (3, 'status', 'active'), "
            "(4, 'username', 'alice'), (4, 'age', '32'), (4, 'height', '185'), (4, 'score', '95'), (4, 'timestamp', '2020-11-01T00:00:00'), (4, 'secret_code', 'abc123'), (4, 'occupation', 'developer'), (4, 'status', 'inactive'), "
            "(5, 'username', 'eve'), (5, 'age', '40'), (5, 'height', '160'), (5, 'score', '85'), (5, 'timestamp', '2021-01-15T00:00:00'), (5, 'secret_code', 'abc123'), (5, 'occupation', 'analyst'), (5, 'status', 'active'), "
            "(6, 'username', 'frank'), (6, 'age', '29'), (6, 'height', '182'), (6, 'score', '105'), (6, 'timestamp', '2020-12-15T00:00:00'), (6, 'secret_code', 'jkl012'), (6, 'occupation', 'developer'), (6, 'status', 'active'), "
            "(7, 'username', 'alice'), (7, 'age', '33'), (7, 'height', '195'), (7, 'score', '88'), (7, 'timestamp', '2020-10-01T00:00:00'), (7, 'secret_code', 'mno345'), (7, 'occupation', 'developer'), (7, 'status', 'active'), "
            "(8, 'username', 'hank'), (8, 'age', '27'), (8, 'height', '165'), (8, 'score', '110'), (8, 'timestamp', '2021-04-01T00:00:00'), (8, 'secret_code', 'pqr678'), (8, 'occupation', 'designer'), (8, 'status', 'inactive'), "
            "(9, 'username', 'alice'), (9, 'age', '36'), (9, 'height', '188'), (9, 'score', '92'), (9, 'timestamp', '2020-09-01T00:00:00'), (9, 'secret_code', 'stu901'), (9, 'occupation', 'developer'), (9, 'status', 'active'), "
            "(10, 'username', 'jack'), (10, 'age', '31'), (10, 'height', '179'), (10, 'score', '115'), (10, 'timestamp', '2021-05-01T00:00:00'), (10, 'secret_code', 'vwx234'), (10, 'occupation', 'teacher'), (10, 'status', 'active'), "
            "(11, 'username', 'kara'), (11, 'age', '26'), (11, 'height', '170'), (11, 'score', '130'), (11, 'timestamp', '2021-06-01T00:00:00'), (11, 'secret_code', 'yza567'), (11, 'occupation', 'developer'), (11, 'status', 'inactive'), "
            "(12, 'username', 'alice'), (12, 'age', '34'), (12, 'height', '192'), (12, 'score', '87'), (12, 'timestamp', '2020-08-01T00:00:00'), (12, 'secret_code', 'bcd890'), (12, 'occupation', 'developer'), (12, 'status', 'active');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print(f"\n-- Expected result: Items {expected_ids}")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(
            sql_query, params, expected_ids, "Complex AND/OR/NOT query"
        )

    def test_empty_query(self):
        query = TagQuery.and_([])
        wql = query.to_wql_str()
        print(f"Test: Empty query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "1=1"
        expected_params = []
        self.assertEqual(sql_query, expected_query, "Empty query mismatch")
        self.assertEqual(params, expected_params, "Empty query params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany("INSERT INTO items (id) VALUES (?)", [(1,), (2,)])
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "value"), (2, "field", "data")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'data');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 2], "Empty query")

    def test_empty_in_list(self):
        query = TagQuery.in_(TagName("field"), [])
        wql = query.to_wql_str()
        print(f"Test: Empty IN list query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value IN ())"
        )
        expected_params = ["field"]
        self.assertEqual(sql_query, expected_query, "Empty IN list query mismatch")
        self.assertEqual(params, expected_params, "Empty IN list params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany("INSERT INTO items (id) VALUES (?)", [(1,), (2,)])
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "field", "value"), (2, "field", "other")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'other');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: No items")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [], "Empty IN list")

    def test_multiple_exists(self):
        query = TagQuery.exist([TagName("f1"), TagName("f2")])
        wql = query.to_wql_str()
        print(f"Test: Multiple EXISTS query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = "(i.id IN (SELECT item_id FROM items_tags WHERE name = ?) AND i.id IN (SELECT item_id FROM items_tags WHERE name = ?))"
        expected_params = ["f1", "f2"]
        self.assertEqual(sql_query, expected_query, "Multiple EXISTS query mismatch")
        self.assertEqual(params, expected_params, "Multiple EXISTS params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "f1", "v1"), (1, "f2", "v2"), (2, "f1", "v1"), (3, "f2", "v2")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'f1', 'v1'), (1, 'f2', 'v2'), "
            "(2, 'f1', 'v1'), "
            "(3, 'f2', 'v2');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 1")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1], "Multiple EXISTS")

    def test_special_characters(self):
        query = TagQuery.eq(TagName("f1"), "val$ue")
        wql = query.to_wql_str()
        print(f"Test: Special characters query\nWQL: {wql}")
        sql_query = self.encoder.encode_query(query)
        params = self.encoder.arguments
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = ? AND value = ?)"
        )
        expected_params = ["f1", "val$ue"]
        self.assertEqual(sql_query, expected_query, "Special characters query mismatch")
        self.assertEqual(params, expected_params, "Special characters params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany("INSERT INTO items (id) VALUES (?)", [(1,), (2,), (3,)])
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (?, ?, ?)",
            [(1, "f1", "val$ue"), (2, "f1", "other"), (3, "f1", "val$ue")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id INTEGER PRIMARY KEY);")
        print("CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT);")
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'f1', 'val$ue'), "
            "(2, 'f1', 'other'), "
            "(3, 'f1', 'val$ue');"
        )
        select_query = f"SELECT id FROM items WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 3")
        print("\n-- Cleanup\nDELETE FROM items_tags; DELETE FROM items;")
        self.run_query_and_verify(sql_query, params, [1, 3], "Special characters")


def main():
    print("Running SqliteTagEncoder non-normalized tests...")
    unittest.main(argv=[""], exit=False)
    print("All tests completed.")


if __name__ == "__main__":
    main()



"""Plugin installer for dynamic plugin installation at runtime."""

import importlib
import json
import logging
import os
import re
import subprocess
import sys
from importlib.metadata import (
    PackageNotFoundError,
    distributions,
)
from importlib.metadata import (
    version as get_package_version,
)
from pathlib import Path
from shutil import which
from subprocess import CompletedProcess
from typing import List, Optional, Set
from urllib.parse import urlparse

from ..version import __version__

LOGGER = logging.getLogger(__name__)

# Valid plugin name pattern: alphanumeric, underscore, hyphen, dot
# Must start with letter or underscore
VALID_PLUGIN_NAME_PATTERN = re.compile(r"^[a-zA-Z_][a-zA-Z0-9_.-]*$")
PLUGIN_REPO_URL = "https://github.com/openwallet-foundation/acapy-plugins"
PYPROJECT_TOML = "pyproject.toml"
PIP_FREEZE_GIT_MARKER = "@ git+"


def _validate_plugin_name(plugin_name: str) -> bool:
    """Validate that a plugin name is safe for use in URLs and file paths.

    Args:
        plugin_name: The plugin name to validate

    Returns:
        True if valid, False otherwise

    """
    if not plugin_name or len(plugin_name) > 100:
        return False
    return bool(VALID_PLUGIN_NAME_PATTERN.match(plugin_name))


def _sanitize_url_component(component: str) -> str:
    """Sanitize a URL component by removing unsafe characters.

    Args:
        component: The URL component to sanitize

    Returns:
        Sanitized component

    """
    # Remove any characters that could be used for URL injection
    return re.sub(r"[^a-zA-Z0-9_.-]", "", component)


def _validate_plugin_source(plugin_source: str) -> bool:
    """Validate that a plugin source URL is safe for use in subprocess calls.

    Args:
        plugin_source: The plugin source URL to validate

    Returns:
        True if valid, False otherwise

    """
    if not plugin_source or len(plugin_source) > 500:
        return False

    # Check for shell injection characters
    dangerous_chars = [";", "&", "|", "`", "$", "(", ")", "<", ">", "\n", "\r"]
    if any(char in plugin_source for char in dangerous_chars):
        return False

    # Validate it's a git+ URL format (expected from _get_plugin_source)
    # Format: git+https://github.com/...@version#subdirectory=...
    # Or allow standard pip package names
    if plugin_source.startswith("git+"):
        # Parse and validate git URL structure
        try:
            # Extract the base URL part (after git+)
            url_with_version = plugin_source[4:]  # Remove "git+"
            url_part = (
                url_with_version.split("@")[0]
                if "@" in url_with_version
                else url_with_version.split("#")[0]
            )
            # Must start with https:// for security
            if url_part.startswith("https://"):
                parsed = urlparse(url_part)
                # Must have a valid netloc (domain)
                if parsed.netloc:
                    return True
        except Exception:
            return False
    elif re.match(r"^[a-zA-Z0-9_.-]+$", plugin_source):
        # Allow simple package names (alphanumeric, dots, hyphens, underscores)
        return True

    return False


def _is_poetry_venv(venv_path: str) -> bool:
    """Check if the virtual environment path indicates Poetry management.

    Args:
        venv_path: Path to the virtual environment

    Returns:
        True if this looks like a Poetry-managed venv

    """
    venv_path_obj = Path(venv_path)
    # Check if this looks like a Poetry-managed venv
    # Poetry venvs are often named like "project-name-<hash>-py3.13"
    if not (venv_path_obj.name.endswith(".venv") or "poetry" in str(venv_path).lower()):
        return False

    # Check if pyproject.toml exists nearby (Poetry projects have it at root)
    parent = venv_path_obj.parent
    return (parent / PYPROJECT_TOML).exists() or (
        venv_path_obj.parent.parent.parent / PYPROJECT_TOML
    ).exists()


def _is_poetry_pyproject(pyproject_file: Path) -> bool:
    """Check if pyproject.toml file indicates Poetry usage.

    Args:
        pyproject_file: Path to pyproject.toml file

    Returns:
        True if pyproject.toml contains Poetry configuration

    """
    if not pyproject_file.exists():
        return False

    try:
        with open(pyproject_file, "r") as f:
            content = f.read()
            return "[tool.poetry]" in content or '[tool."poetry.core"]' in content
    except Exception:
        return False


def _get_pyproject_search_paths() -> List[Path]:
    """Get list of paths to search for pyproject.toml.

    Returns:
        List of directory paths to check

    """
    search_paths = [Path.cwd()]

    # Also check from the acapy_agent module location (project root)
    try:
        from .. import __file__ as module_file

        module_path = Path(module_file).resolve()
        # Go up from acapy_agent/utils/plugin_installer.py to project root
        project_root = module_path.parent.parent.parent
        if project_root not in search_paths:
            search_paths.append(project_root)
    except Exception:
        # It is safe to ignore errors here; failure to import the module or
        # resolve its path simply means we cannot add an extra search path
        # for pyproject.toml detection.
        pass

    return search_paths


def _detect_package_manager() -> Optional[str]:
    """Detect which package manager is being used (poetry, pip, etc.).

    Returns:
        "poetry" if Poetry is detected, None otherwise

    """
    # Check if poetry is available
    if not which("poetry"):
        return None

    # Check if we're in a Poetry-managed virtual environment
    # Poetry typically sets VIRTUAL_ENV to a path containing ".venv" or in Poetry's cache
    venv_path = os.environ.get("VIRTUAL_ENV")
    if venv_path and _is_poetry_venv(venv_path):
        return "poetry"

    # Check if we're in a Poetry project by looking for pyproject.toml
    for root_path in _get_pyproject_search_paths():
        if _is_poetry_pyproject(root_path / PYPROJECT_TOML):
            return "poetry"

    return None


def _get_pip_command_base() -> List[str]:
    """Get the appropriate pip command base for the current environment.

    Returns:
        List of command parts to run pip
        (e.g., ["poetry", "run", "pip"] or [sys.executable, "-m", "pip"])

    """
    package_manager = _detect_package_manager()
    if package_manager == "poetry":
        # Use poetry run pip in Poetry environments
        return ["poetry", "run", "pip"]
    else:
        # Use sys.executable -m pip for regular Python environments
        return [sys.executable, "-m", "pip"]


def _get_pip_command() -> List[str]:
    """Get the appropriate pip install command for the current environment.

    Returns:
        List of command parts to run pip install

    """
    cmd = _get_pip_command_base()
    cmd.append("install")
    return cmd


class PluginInstaller:
    """Handles dynamic installation of ACA-Py plugins from acapy-plugins."""

    def __init__(
        self,
        auto_install: bool = True,
        plugin_version: Optional[str] = None,
    ):
        """Initialize the plugin installer.

        Args:
            auto_install: Whether to automatically install missing plugins
            plugin_version: Version to use for plugin installation.
                If None, uses current ACA-Py version.

        """
        self.auto_install = auto_install
        self.plugin_version = plugin_version
        self.installed_plugins: Set[str] = set()

    def _try_get_package_version(
        self, names: List[str]
    ) -> tuple[Optional[str], Optional[str]]:
        """Try to get package version from multiple name variations.

        Returns:
            (version, package_name) tuple

        """
        for name in names:
            try:
                return get_package_version(name), name
            except PackageNotFoundError:
                continue
        return None, None

    def _is_valid_revision(self, revision: str) -> bool:
        """Check if a revision/tag string is valid.

        Args:
            revision: The revision or tag to validate

        Returns:
            True if revision is valid

        """
        if not revision:
            return False
        return (
            "." in revision or revision in ["main", "master", "develop"]
        ) and re.match(r"^[a-zA-Z0-9._-]+$", revision)

    def _extract_tag_from_path(self, parsed_url) -> Optional[str]:
        """Extract tag from URL path (standard format).

        Args:
            parsed_url: Parsed URL object

        Returns:
            Tag string if found and valid, None otherwise

        """
        if "@" not in parsed_url.path:
            return None

        path_parts = parsed_url.path.rsplit("@", 1)
        if len(path_parts) != 2:
            return None

        tag = path_parts[1].split("#")[0]  # Remove fragment if present
        return tag if self._is_valid_revision(tag) else None

    def _extract_tag_from_netloc(self, parsed_url) -> Optional[str]:
        """Extract tag from URL netloc (non-standard format).

        Args:
            parsed_url: Parsed URL object

        Returns:
            Tag string if found and valid, None otherwise

        """
        if not parsed_url.scheme or "@" not in parsed_url.netloc:
            return None

        netloc_parts = parsed_url.netloc.rsplit("@", 1)
        if len(netloc_parts) != 2:
            return None

        tag = netloc_parts[1]
        return tag if self._is_valid_revision(tag) else None

    def _extract_version_from_url(self, url: str) -> Optional[str]:
        """Extract version tag from a Git URL.

        Args:
            url: The Git URL to parse

        Returns:
            Version tag if found, None otherwise

        """
        try:
            parsed = urlparse(url)
            # Try standard format first: @version in path
            tag = self._extract_tag_from_path(parsed)
            if tag:
                return tag

            # Fallback: non-standard format with @version in netloc
            tag = self._extract_tag_from_netloc(parsed)
            if tag:
                return tag
        except Exception:
            LOGGER.debug("Failed to parse URL: %s", url)
        return None

    def _extract_source_version_from_direct_url(
        self, direct_url_data: dict
    ) -> Optional[str]:
        """Extract git tag/version from direct_url.json data."""
        # Try vcs_info first (primary method)
        vcs_info = direct_url_data.get("vcs_info", {})
        if vcs_info.get("vcs") == "git":
            revision = vcs_info.get("requested_revision")
            if self._is_valid_revision(revision):
                return revision

        # Fallback: Try to extract version from URL if vcs_info didn't work
        # This handles edge cases where vcs_info.requested_revision might be missing
        url = direct_url_data.get("url", "")
        if url:
            return self._extract_version_from_url(url)

        return None

    def _get_location_from_pip_show(self, package_name: str) -> Optional[str]:
        """Get package location using pip show.

        Args:
            package_name: Name of the package

        Returns:
            Location path if found, None otherwise

        """
        try:
            cmd = _get_pip_command_base()
            cmd.extend(["show", package_name])
            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                check=False,
            )
            if result.returncode != 0:
                return None

            return next(
                (
                    line.split(":", 1)[1].strip()
                    for line in result.stdout.split("\n")
                    if line.startswith("Location:")
                ),
                None,
            )
        except Exception as e:
            LOGGER.exception(
                f"Error while trying to locate package '{package_name}': {e}"
            )
            return None

    def _read_direct_url_file(self, direct_url_file: Path) -> Optional[str]:
        """Read and extract version from a direct_url.json file.

        Args:
            direct_url_file: Path to direct_url.json file

        Returns:
            Source version if found, None otherwise

        """
        if not direct_url_file.exists():
            return None

        try:
            with open(direct_url_file) as f:
                source_version = self._extract_source_version_from_direct_url(
                    json.load(f)
                )
                return source_version
        except (json.JSONDecodeError, IOError) as e:
            LOGGER.debug(
                "Failed to read or parse direct_url.json for %s: %s",
                direct_url_file,
                e,
            )
            return None

    def _find_direct_url_in_location(self, location: str) -> Optional[str]:
        """Find and read direct_url.json in a package location.

        Args:
            location: Package installation location

        Returns:
            Source version if found, None otherwise

        """
        try:
            location_path = Path(location)
            for item in location_path.iterdir():
                if item.is_dir() and item.name.endswith(".dist-info"):
                    direct_url_file = item / "direct_url.json"
                    source_version = self._read_direct_url_file(direct_url_file)
                    if source_version:
                        return source_version
        except Exception as e:
            LOGGER.warning(
                "Failed to search location '%s' for direct_url.json: %s",
                location,
                e,
            )
        return None

    def _search_distributions_for_direct_url(self, package_name: str) -> Optional[str]:
        """Search installed distributions for direct_url.json.

        Args:
            package_name: Name of the package to search for

        Returns:
            Source version if found, None otherwise

        """
        for dist in distributions():
            if dist.metadata["Name"].lower() != package_name.lower():
                continue

            dist_location = Path(dist.location)
            pkg_name, pkg_version = dist.metadata["Name"], dist.version
            name_variants = [
                f"{pkg_name}-{pkg_version}.dist-info",
                f"{pkg_name.replace('-', '_')}-{pkg_version}.dist-info",
                f"{pkg_name.replace('.', '_')}-{pkg_version}.dist-info",
            ]

            for name_variant in name_variants:
                direct_url_file = dist_location / name_variant / "direct_url.json"
                source_version = self._read_direct_url_file(direct_url_file)
                if source_version:
                    return source_version

        return None

    def _extract_version_from_pip_freeze(self, package_name: str) -> Optional[str]:
        """Extract version from pip freeze output.

        Args:
            package_name: Name of the package

        Returns:
            Source version if found, None otherwise

        """
        try:
            cmd = _get_pip_command_base()
            cmd.append("freeze")
            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                check=False,
            )
            if result.returncode != 0:
                return None

            for line in result.stdout.split("\n"):
                if (
                    package_name.lower() not in line.lower()
                    or PIP_FREEZE_GIT_MARKER not in line
                ):
                    continue

                # Extract git URL from pip freeze line
                # Format: package==version @ git+https://github.com/org/repo@tag#subdirectory=...
                try:
                    git_url_part = line.split(PIP_FREEZE_GIT_MARKER, 1)[1]
                    parsed = urlparse(f"git+{git_url_part}")
                    # Try standard format first, then fallback
                    tag = self._extract_tag_from_path(parsed)
                    if tag:
                        return tag

                    tag = self._extract_tag_from_netloc(parsed)
                    if tag:
                        return tag
                except Exception:
                    LOGGER.debug("Failed to parse git URL from pip freeze line: %s", line)
                    continue
        except Exception as e:
            LOGGER.debug(
                "Exception occurred while running pip freeze to get source version "
                "for %s: %s",
                package_name,
                e,
                exc_info=True,
            )
        return None

    def _get_source_version_from_dist_info(self, package_name: str) -> Optional[str]:
        """Get source version from pip's .dist-info/direct_url.json file."""
        # Strategy 1: Use pip show to find location
        location = self._get_location_from_pip_show(package_name)
        if location:
            source_version = self._find_direct_url_in_location(location)
            if source_version:
                return source_version

        # Strategy 2: Search distributions
        source_version = self._search_distributions_for_direct_url(package_name)
        if source_version:
            return source_version

        # Strategy 3: Last resort - pip freeze
        return self._extract_version_from_pip_freeze(package_name)

    def _get_installed_plugin_version(self, plugin_name: str) -> Optional[dict]:
        """Get version info of an installed plugin.

        Returns package version and source version (git tag) if available.
        """
        result = {}

        # Try to get package version from various name variations
        name_variants = [
            plugin_name,
            plugin_name.replace("_", "-"),
            f"acapy-plugin-{plugin_name.replace('_', '-')}",
        ]
        package_version, package_name = self._try_get_package_version(name_variants)

        if not package_version:
            # Try __version__ attribute
            # Note: We avoid importing the module if possible to prevent side effects
            # Only try this as a last resort
            try:
                # Check if module is already loaded before importing
                if plugin_name in sys.modules:
                    module = sys.modules[plugin_name]
                else:
                    # Only import if not already loaded to avoid side effects
                    module = importlib.import_module(plugin_name)
                if hasattr(module, "__version__"):
                    package_version = str(module.__version__)
            except (ImportError, AttributeError):
                # Catch import/attribute errors to prevent side effects from breaking
                # version lookup
                pass

        if not package_version:
            return None

        result["package_version"] = package_version

        # Try to get source version (git tag) if we found a package name
        if package_name:
            source_version = self._get_source_version_from_dist_info(package_name)
            if source_version:
                result["source_version"] = source_version

        return result

    def _get_plugin_source(self, plugin_name: str) -> str:
        """Get the installation source for a plugin from acapy-plugins repository.

        Args:
            plugin_name: The plugin name (must be validated before calling)

        Returns:
            Git URL for installing the plugin

        Raises:
            ValueError: If plugin_name is invalid or unsafe

        """
        # Validate plugin name to prevent URL injection
        if not _validate_plugin_name(plugin_name):
            raise ValueError(
                f"Invalid plugin name: '{plugin_name}'. "
                "Plugin names must contain only alphanumeric characters, "
                "underscores, hyphens, and dots, and must start with a letter "
                "or underscore."
            )

        # Sanitize version if provided
        version_to_use = (
            self.plugin_version if self.plugin_version is not None else __version__
        )
        version_to_use = _sanitize_url_component(str(version_to_use))

        # Sanitize plugin name (though it should already be valid)
        sanitized_plugin_name = _sanitize_url_component(plugin_name)

        # Construct URL with validated and sanitized components
        return (
            f"git+{PLUGIN_REPO_URL}@{version_to_use}#subdirectory={sanitized_plugin_name}"
        )

    def _extract_version_info(self, plugin_source: str) -> str:
        """Extract version information from plugin source URL for logging.

        Args:
            plugin_source: The plugin source URL

        Returns:
            Version info string (e.g., " (version: 1.0.0)") or empty string

        """
        if "@" not in plugin_source:
            return ""

        parts = plugin_source.split("@")
        if len(parts) <= 1:
            return ""

        version_part = parts[1].split("#")[0] if "#" in parts[1] else parts[1]
        return f" (version: {version_part})"

    def _log_installation_start(
        self, plugin_name: str, plugin_source: str, version_info: str, upgrade: bool
    ):
        """Log the start of plugin installation.

        Args:
            plugin_name: The plugin name
            plugin_source: The plugin source URL
            version_info: Version information string
            upgrade: Whether this is an upgrade

        """
        display_name = plugin_name or plugin_source
        if upgrade:
            LOGGER.info("Upgrading plugin '%s'%s", display_name, version_info)
        else:
            LOGGER.info("Installing plugin '%s'%s", display_name, version_info)

    def _build_install_command(self, plugin_source: str, upgrade: bool) -> List[str]:
        """Build the pip install command.

        Args:
            plugin_source: The plugin source URL
            upgrade: Whether to upgrade the plugin

        Returns:
            List of command parts

        """
        cmd = _get_pip_command()
        cmd.extend(["--no-cache-dir"])
        if upgrade:
            cmd.extend(["--upgrade", "--force-reinstall", "--no-deps"])
        cmd.append(plugin_source)
        return cmd

    def _handle_install_result(
        self,
        result: CompletedProcess,
        plugin_name: str,
        plugin_source: str,
        version_info: str,
        upgrade: bool,
    ) -> bool:
        """Handle the result of plugin installation.

        Args:
            result: The subprocess result
            plugin_name: The plugin name
            plugin_source: The plugin source URL
            version_info: Version information string
            upgrade: Whether this was an upgrade

        Returns:
            True if installation succeeded, False otherwise

        """
        display_name = plugin_name or plugin_source

        if result.returncode == 0:
            action = "Upgraded" if upgrade else "Installed"
            LOGGER.info(
                "Successfully %s plugin '%s'%s",
                action.lower(),
                display_name,
                version_info,
            )
            return True

        action = "upgrade" if upgrade else "install"
        LOGGER.error(
            "Failed to %s plugin '%s'%s: %s",
            action,
            display_name,
            version_info,
            result.stderr,
        )
        return False

    def _install_plugin(
        self, plugin_source: str, plugin_name: str, upgrade: bool = False
    ) -> bool:
        """Install a plugin using pip or poetry run pip.

        Args:
            plugin_source: The plugin source URL (should come from _get_plugin_source)
            plugin_name: The plugin name
            upgrade: Whether to upgrade the plugin

        Returns:
            True if installation succeeded, False otherwise

        Raises:
            ValueError: If plugin_source is invalid or unsafe

        """
        # Validate plugin_source to prevent command injection
        if not _validate_plugin_source(plugin_source):
            raise ValueError(
                f"Invalid or unsafe plugin_source: '{plugin_source}'. "
                "Plugin sources must be valid git+ URLs or safe package names. "
                "Use _get_plugin_source() to generate safe plugin sources."
            )

        try:
            version_info = self._extract_version_info(plugin_source)
            self._log_installation_start(
                plugin_name, plugin_source, version_info, upgrade
            )

            cmd = self._build_install_command(plugin_source, upgrade)
            result = subprocess.run(cmd, capture_output=True, text=True, check=False)

            return self._handle_install_result(
                result, plugin_name, plugin_source, version_info, upgrade
            )
        except Exception as e:
            LOGGER.error(
                "Error installing plugin %s: %s", plugin_name or plugin_source, e
            )
            return False

    def _check_plugin_exists(self, plugin_name: str) -> bool:
        """Check if a plugin can be imported (exists).

        Args:
            plugin_name: The name of the plugin module

        Returns:
            True if plugin can be imported, False otherwise

        """
        try:
            importlib.import_module(plugin_name)
            return True
        except ImportError:
            return False

    def _get_installed_version_info(
        self, plugin_name: str
    ) -> tuple[Optional[str], Optional[str]]:
        """Get installed version information for a plugin.

        Args:
            plugin_name: The name of the plugin module

        Returns:
            Tuple of (package_version, source_version)

        """
        version_info = self._get_installed_plugin_version(plugin_name)
        if not version_info:
            return None, None

        return (
            version_info.get("package_version"),
            version_info.get("source_version"),
        )

    def _normalize_version(self, version: str) -> str:
        """Normalize version string by removing build/metadata suffixes.

        Args:
            version: Version string to normalize

        Returns:
            Normalized version string

        """
        return version.split("+")[0].split("-")[0].strip()

    def _check_version_matches_implicit(
        self,
        plugin_name: str,
        installed_package_version: Optional[str],
        target_version: str,
    ) -> bool:
        """Check if installed version matches target when no explicit version specified.

        Args:
            plugin_name: The name of the plugin module
            installed_package_version: Installed package version
            target_version: Target version to match

        Returns:
            True if versions match, False otherwise

        """
        if not installed_package_version:
            return False

        normalized_installed = self._normalize_version(installed_package_version)
        normalized_target = self._normalize_version(target_version)

        if normalized_installed == normalized_target:
            LOGGER.debug(
                "Plugin '%s' already installed with matching version: %s",
                plugin_name,
                installed_package_version,
            )
            return True

        return False

    def _check_version_matches_explicit(
        self,
        plugin_name: str,
        installed_source_version: Optional[str],
        installed_package_version: Optional[str],
        target_version: str,
    ) -> bool:
        """Check if installed version matches target when explicit version specified.

        Args:
            plugin_name: The name of the plugin module
            installed_source_version: Installed source version (git tag)
            installed_package_version: Installed package version
            target_version: Target version to match

        Returns:
            True if versions match, False otherwise

        """
        if installed_source_version and installed_source_version == target_version:
            LOGGER.info(
                "Plugin '%s' already installed from source version %s "
                "(package version: %s)",
                plugin_name,
                installed_source_version,
                installed_package_version or "unknown",
            )
            return True

        return False

    def _log_version_mismatch(
        self,
        plugin_name: str,
        installed_source_version: Optional[str],
        installed_package_version: Optional[str],
        target_version: str,
    ):
        """Log version mismatch details.

        Args:
            plugin_name: The name of the plugin module
            installed_source_version: Installed source version (git tag)
            installed_package_version: Installed package version
            target_version: Target version

        """
        if installed_source_version:
            LOGGER.info(
                "Plugin '%s' source version mismatch detected: "
                "installed=%s, target=%s. Upgrading plugin...",
                plugin_name,
                installed_source_version,
                target_version,
            )
        elif installed_package_version:
            LOGGER.info(
                "Plugin '%s' needs upgrade: current package version=%s, "
                "target version=%s. Upgrading plugin...",
                plugin_name,
                installed_package_version,
                target_version,
            )
        else:
            LOGGER.info(
                "Plugin '%s' is installed but version cannot be determined. "
                "Upgrading to ensure correct version (%s)...",
                plugin_name,
                target_version,
            )

    def _verify_plugin_import(self, plugin_name: str) -> bool:
        """Verify that an installed plugin can be imported.

        Args:
            plugin_name: The name of the plugin module

        Returns:
            True if plugin can be imported, False otherwise

        """
        try:
            importlib.import_module(plugin_name)
            return True
        except ImportError as e:
            LOGGER.error(
                "Plugin '%s' was installed but cannot be imported: %s",
                plugin_name,
                e,
            )
            return False

    def _check_and_handle_existing_plugin(
        self,
        plugin_name: str,
        installed_package_version: Optional[str],
        installed_source_version: Optional[str],
        target_version: str,
    ) -> bool:
        """Check if existing plugin version matches and handle accordingly.

        Args:
            plugin_name: The name of the plugin module
            installed_package_version: Installed package version
            installed_source_version: Installed source version
            target_version: Target version to match

        Returns:
            True if plugin is already correctly installed, False if needs installation

        """
        if not self.plugin_version:
            # No explicit version - check package version match
            if self._check_version_matches_implicit(
                plugin_name, installed_package_version, target_version
            ):
                self.installed_plugins.add(plugin_name)
                return True
            LOGGER.info(
                "Plugin '%s' exists but version check inconclusive. "
                "Reinstalling to ensure correct version (%s)...",
                plugin_name,
                target_version,
            )
        else:
            # Explicit version - check source version match
            if self._check_version_matches_explicit(
                plugin_name,
                installed_source_version,
                installed_package_version,
                target_version,
            ):
                self.installed_plugins.add(plugin_name)
                return True
            self._log_version_mismatch(
                plugin_name,
                installed_source_version,
                installed_package_version,
                target_version,
            )
        return False

    def _attempt_plugin_installation(
        self, plugin_name: str, plugin_exists: bool, target_version: str
    ) -> bool:
        """Attempt to install or upgrade a plugin.

        Args:
            plugin_name: The name of the plugin module
            plugin_exists: Whether plugin already exists
            target_version: Target version

        Returns:
            True if installation succeeded, False otherwise

        """
        if not self.auto_install:
            LOGGER.warning(
                "Plugin '%s' is not installed and auto-install is disabled", plugin_name
            )
            return False

        plugin_source = self._get_plugin_source(plugin_name)
        if self._install_plugin(
            plugin_source, plugin_name=plugin_name, upgrade=plugin_exists
        ) and self._verify_plugin_import(plugin_name):
            self.installed_plugins.add(plugin_name)
            return True

        LOGGER.error(
            "Failed to install plugin '%s' (version %s)",
            plugin_name,
            target_version,
        )
        return False

    def ensure_plugin_installed(self, plugin_name: str) -> bool:
        """Ensure a plugin is installed with the correct version.

        If not installed or version doesn't match, attempt to install it.

        Args:
            plugin_name: The name of the plugin module (must be validated)

        Returns:
            True if plugin is available (was already installed or successfully installed)

        Raises:
            ValueError: If plugin_name is invalid or unsafe

        """
        # Validate plugin name before processing
        if not _validate_plugin_name(plugin_name):
            raise ValueError(
                f"Invalid plugin name: '{plugin_name}'. "
                "Plugin names must contain only alphanumeric characters, "
                "underscores, hyphens, and dots, and must start with a letter "
                "or underscore."
            )

        target_version = (
            self.plugin_version if self.plugin_version is not None else __version__
        )

        plugin_exists = self._check_plugin_exists(plugin_name)
        installed_package_version, installed_source_version = (
            self._get_installed_version_info(plugin_name)
            if plugin_exists
            else (None, None)
        )

        # Check if version matches (different logic for explicit vs implicit versions)
        if plugin_exists:
            if self._check_and_handle_existing_plugin(
                plugin_name,
                installed_package_version,
                installed_source_version,
                target_version,
            ):
                return True
        else:
            LOGGER.info(
                "Plugin '%s' not found. Installing version %s...",
                plugin_name,
                target_version,
            )

        return self._attempt_plugin_installation(
            plugin_name, plugin_exists, target_version
        )

    def ensure_plugins_installed(self, plugin_names: List[str]) -> List[str]:
        """Ensure multiple plugins are installed.

        Args:
            plugin_names: List of plugin module names

        Returns:
            List of plugin names that failed to install

        """
        failed = []
        for plugin_name in plugin_names:
            if not self.ensure_plugin_installed(plugin_name):
                failed.append(plugin_name)

        return failed


def install_plugins_from_config(
    plugin_names: List[str],
    auto_install: bool = True,
    plugin_version: Optional[str] = None,
) -> List[str]:
    """Install plugins from a list of plugin names.

    Args:
        plugin_names: List of plugin module names to install
        auto_install: Whether to automatically install missing plugins
        plugin_version: Version to use for plugin installation.
            If None, uses current ACA-Py version.

    Returns:
        List of plugin names that failed to install

    """
    if not plugin_names:
        return []

    installer = PluginInstaller(
        auto_install=auto_install,
        plugin_version=plugin_version,
    )

    return installer.ensure_plugins_installed(plugin_names)


def get_plugin_version(plugin_name: str) -> Optional[dict]:
    """Get the installed version information of a plugin.

    Includes package version and installation source.

    Args:
        plugin_name: The name of the plugin module

    Returns:
        Dictionary with 'package_version' and optionally 'source_version'
        (git tag), or None if not found. Returns None if any error occurs.

    """
    try:
        installer = PluginInstaller(auto_install=False)
        return installer._get_installed_plugin_version(plugin_name)
    except Exception:
        # Silently fail version lookup - don't break plugin functionality
        LOGGER.debug(
            "Failed to get version info for plugin '%s'", plugin_name, exc_info=True
        )
        return None


def list_plugin_versions(plugin_names: List[str] = None) -> dict:
    """Get version information for a list of plugins, or all installed plugins.

    Args:
        plugin_names: Optional list of plugin names to check.
            If None, attempts to detect installed plugins.

    Returns:
        Dictionary mapping plugin names to their version info dicts
        (or None if version cannot be determined)

    """
    installer = PluginInstaller(auto_install=False)
    result = {}

    if plugin_names:
        for plugin_name in plugin_names:
            version_info = installer._get_installed_plugin_version(plugin_name)
            result[plugin_name] = version_info
    else:
        # Try to detect installed plugins - limited without knowing what's
        # installed
        # For now, just return empty dict - callers should provide plugin names
        pass

    return result



import os

import pytest
import pytest_asyncio

from ...core.profile import Profile
from ...ledger.base import BaseLedger
from ...tests import mock
from ...utils.testing import create_test_profile
from ..did_method import INDY, SOV, WEB
from ..error import WalletDuplicateError, WalletError, WalletNotFoundError
from ..kanon_wallet import KanonWallet
from ..key_type import ED25519

# Skip all tests if POSTGRES_URL is not set
if not os.getenv("POSTGRES_URL"):
    pytest.skip(
        "Kanon PostgreSQL integration tests disabled: set POSTGRES_URL to enable",
        allow_module_level=True,
    )

pytestmark = [pytest.mark.postgres, pytest.mark.p1]


@pytest_asyncio.fixture
async def profile():
    postgres_url = os.getenv("POSTGRES_URL")
    profile = await create_test_profile(
        settings={
            "wallet.type": "kanon-anoncreds",
            "wallet.storage_type": "postgres",
            "wallet.storage_config": {"url": postgres_url},
            "wallet.storage_creds": {
                "account": "postgres",
                "password": "postgres",
            },
            "dbstore.storage_type": "postgres",
            "dbstore.storage_config": {"url": postgres_url},
            "dbstore.storage_creds": {
                "account": "postgres",
                "password": "postgres",
            },
            "dbstore.schema_config": "normalize",
        }
    )
    yield profile
    # Cleanup happens automatically


@pytest_asyncio.fixture
async def wallet(profile: Profile):
    async with profile.session() as session:
        yield KanonWallet(session)


@pytest.mark.asyncio
async def test_create_local_did(wallet: KanonWallet):
    metadata = {"description": "Test DID", "public": False}
    did_info = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
        metadata=metadata,
    )

    assert did_info.did
    assert did_info.verkey
    assert did_info.metadata == metadata
    assert did_info.method == SOV
    assert did_info.key_type == ED25519

    retrieved = await wallet.get_local_did(did_info.did)
    assert retrieved.did == did_info.did
    assert retrieved.verkey == did_info.verkey
    assert retrieved.metadata == metadata


@pytest.mark.asyncio
async def test_create_public_did(wallet: KanonWallet):
    did_info = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
        metadata={"public": True},
    )

    await wallet.set_public_did(did_info.did)

    public_did = await wallet.get_public_did()
    assert public_did is not None
    assert public_did.did == did_info.did
    assert public_did.verkey == did_info.verkey


@pytest.mark.asyncio
async def test_rotate_keypair(wallet: KanonWallet):
    did_info = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
    )
    old_verkey = did_info.verkey

    new_verkey = await wallet.rotate_did_keypair_start(did_info.did)
    await wallet.rotate_did_keypair_apply(did_info.did)

    assert new_verkey != old_verkey

    updated_did = await wallet.get_local_did(did_info.did)
    assert updated_did.verkey == new_verkey

    message = b"test message after rotation"
    signature = await wallet.sign_message(message, new_verkey)
    assert signature

    assert updated_did.verkey != old_verkey


@pytest.mark.asyncio
async def test_get_all_local_dids(wallet: KanonWallet):
    did_infos = []
    for i in range(3):
        did_info = await wallet.create_local_did(
            method=SOV,
            key_type=ED25519,
            metadata={"index": i, "description": f"Test DID {i}"},
        )
        did_infos.append(did_info)

    all_dids = await wallet.get_local_dids()

    created_dids = {di.did for di in did_infos}
    retrieved_dids = {di.did for di in all_dids}
    assert created_dids.issubset(retrieved_dids)

    test_did = next(di for di in all_dids if di.did == did_infos[0].did)
    assert test_did.verkey == did_infos[0].verkey
    assert test_did.metadata.get("index") == 0


@pytest.mark.asyncio
async def test_sign_message(wallet: KanonWallet):
    did_info = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
    )

    message = b"Hello, World!"
    signature = await wallet.sign_message(message, did_info.verkey)

    assert signature
    assert isinstance(signature, bytes)

    valid = await wallet.verify_message(message, signature, did_info.verkey, ED25519)
    assert valid

    message2 = b"Different message"
    signature2 = await wallet.sign_message(message2, did_info.verkey)
    assert signature2 != signature


@pytest.mark.asyncio
async def test_verify_message(wallet: KanonWallet):
    did_info = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
    )

    message = b"Test message for verification"
    signature = await wallet.sign_message(message, did_info.verkey)

    valid = await wallet.verify_message(message, signature, did_info.verkey, ED25519)
    assert valid is True

    invalid_sig = b"invalid signature bytes"
    valid = await wallet.verify_message(message, invalid_sig, did_info.verkey, ED25519)
    assert valid is False

    wrong_message = b"Wrong message content"
    valid = await wallet.verify_message(
        wrong_message, signature, did_info.verkey, ED25519
    )
    assert valid is False


@pytest.mark.asyncio
async def test_pack_message(wallet: KanonWallet):
    sender = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
    )
    recipient = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
    )

    message = b"Secret message content"
    packed = await wallet.pack_message(message, [recipient.verkey], sender.verkey)

    assert packed
    assert isinstance(packed, bytes)

    assert packed != message

    assert len(packed) > len(message)


@pytest.mark.asyncio
async def test_unpack_message(wallet: KanonWallet):
    sender = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
    )
    recipient = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
    )

    original_message = b"Test message for pack/unpack"
    packed = await wallet.pack_message(
        original_message, [recipient.verkey], sender.verkey
    )
    message, from_verkey, to_verkey = await wallet.unpack_message(packed)

    assert message == original_message.decode("utf-8")
    assert from_verkey == sender.verkey
    assert to_verkey == recipient.verkey


@pytest.mark.asyncio
async def test_set_public_did(wallet: KanonWallet):
    did1 = await wallet.create_local_did(method=SOV, key_type=ED25519)
    did2 = await wallet.create_local_did(method=SOV, key_type=ED25519)

    await wallet.set_public_did(did1.did)
    public = await wallet.get_public_did()
    assert public.did == did1.did

    await wallet.set_public_did(did2.did)
    public = await wallet.get_public_did()
    assert public.did == did2.did

    assert public.did != did1.did


@pytest.mark.asyncio
async def test_replace_local_did_metadata(wallet: KanonWallet):
    initial_metadata = {"description": "Initial", "version": 1}
    did_info = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
        metadata=initial_metadata,
    )

    retrieved = await wallet.get_local_did(did_info.did)
    assert retrieved.metadata == initial_metadata

    new_metadata = {"description": "Updated", "version": 2, "active": True}
    await wallet.replace_local_did_metadata(did_info.did, new_metadata)

    updated = await wallet.get_local_did(did_info.did)
    assert updated.metadata == new_metadata
    assert updated.metadata != initial_metadata
    assert updated.metadata["version"] == 2


@pytest.mark.asyncio
async def test_get_local_did_by_verkey(wallet: KanonWallet):
    metadata = {"description": "Test for verkey lookup"}
    did_info = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
        metadata=metadata,
    )

    retrieved = await wallet.get_local_did_for_verkey(did_info.verkey)

    assert retrieved.did == did_info.did
    assert retrieved.verkey == did_info.verkey
    assert retrieved.metadata == metadata

    with pytest.raises(WalletNotFoundError):
        await wallet.get_local_did_for_verkey("NonExistentVerkey123")


@pytest.mark.asyncio
async def test_set_did_endpoint(wallet: KanonWallet):
    mock_ledger = mock.MagicMock(BaseLedger, autospec=True)

    sov_did_info = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
    )

    await wallet.set_public_did(sov_did_info.did)

    original_replace = wallet.replace_local_did_metadata
    wallet.replace_local_did_metadata = mock.CoroutineMock()

    await wallet.set_did_endpoint(
        sov_did_info.did,
        "http://example.com",
        mock_ledger,
    )

    wallet.replace_local_did_metadata = original_replace

    indy_did_info = await wallet.create_local_did(
        method=INDY,
        key_type=ED25519,
    )

    await wallet.set_public_did(indy_did_info.did)

    wallet.replace_local_did_metadata = mock.CoroutineMock()
    await wallet.set_did_endpoint(
        indy_did_info.did,
        "http://example.com",
        mock_ledger,
    )
    wallet.replace_local_did_metadata = original_replace

    import uuid

    web_did_info = await wallet.create_local_did(
        method=WEB,
        key_type=ED25519,
        did=f"did:web:example.com:test-{uuid.uuid4()}",
    )

    await wallet.set_public_did(web_did_info.did)

    with pytest.raises(WalletError):
        await wallet.set_did_endpoint(
            web_did_info.did,
            "http://example.com",
            mock_ledger,
        )


@pytest.mark.asyncio
async def test_duplicate_did_error(wallet: KanonWallet):
    seed = "000000000000000000000000Wallet01"
    did_info1 = await wallet.create_local_did(
        method=SOV,
        key_type=ED25519,
        seed=seed,
        metadata={"original": True},
    )

    try:
        did_info2 = await wallet.create_local_did(
            method=SOV,
            key_type=ED25519,
            seed=seed,
            metadata={"original": False},
        )
        assert did_info2.did == did_info1.did
    except WalletDuplicateError:
        pass

    retrieved = await wallet.get_local_did(did_info1.did)
    assert retrieved.did == did_info1.did


@pytest.mark.asyncio
async def test_get_nonexistent_did_error(wallet: KanonWallet):
    with pytest.raises(WalletNotFoundError):
        await wallet.get_local_did("NonExistentDID123456")

    with pytest.raises(WalletNotFoundError):
        await wallet.get_local_did_for_verkey("NonExistentVerkey123456")



"""Common test fixtures and data for AnonCreds routes."""

from typing import Any

from ....admin.request_context import AdminRequestContext
from ....tests import mock
from ....utils.testing import create_test_profile
from ...models.revocation import RevRegDef, RevRegDefValue


class BaseAnonCredsRouteTestCase:
    """Base test case with common setup for AnonCreds route tests."""

    async def asyncSetUp(self) -> None:
        """Common test setup for all AnonCreds route tests."""
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "wallet.type": "askar-anoncreds",
                "admin.admin_api_key": "secret-key",
            },
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )

        self.test_did = "sample-did"


class BaseAnonCredsRouteTestCaseWithOutbound(BaseAnonCredsRouteTestCase):
    """Base test case with outbound message router for AnonCreds route tests."""

    async def asyncSetUp(self) -> None:
        """Common test setup with outbound message router."""
        await super().asyncSetUp()
        self.request_dict["outbound_message_router"] = mock.CoroutineMock()

        # Update the mock's __getitem__ behavior to handle the new key
        def getitem_side_effect(obj: Any, key: str) -> Any:
            return self.request_dict[key]

        self.request.__getitem__.side_effect = getitem_side_effect


def create_mock_request(context: AdminRequestContext, **kwargs) -> mock.MagicMock:
    """Create a mock request object for testing."""
    request_dict = {"context": context}
    request_dict.update(kwargs)

    return mock.MagicMock(
        app={},
        match_info={},
        query={},
        __getitem__=lambda _, k: request_dict[k],
        context=context,
        headers={"x-api-key": "secret-key"},
    )


def create_standard_rev_reg_def(
    tag: str = "tag",
    cred_def_id: str = "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
    issuer_id: str = "CsQY9MGeD3CQP4EyuVFo5m",
    max_cred_num: int = 100,
) -> RevRegDef:
    """Create a standard revocation registry definition for testing."""
    return RevRegDef(
        tag=tag,
        cred_def_id=cred_def_id,
        value=RevRegDefValue(
            max_cred_num=max_cred_num,
            public_keys={
                "accum_key": {"z": "1 0BB...386"},
            },
            tails_hash="58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
            tails_location="http://tails-server.com",
        ),
        issuer_id=issuer_id,
        type="CL_ACCUM",
    )


def create_standard_rev_reg_def_value(
    max_cred_num: int = 100,
    tails_hash: str = "58NNWYnVxVFzAfUztwGSNBL4551XNq6nXk56pCiKJxxt",
    tails_location: str = "http://tails-server.com",
) -> RevRegDefValue:
    """Create a standard revocation registry definition value for testing."""
    return RevRegDefValue(
        max_cred_num=max_cred_num,
        public_keys={
            "accum_key": {"z": "1 0BB...386"},
        },
        tails_hash=tails_hash,
        tails_location=tails_location,
    )



"""Askar store configuration and management."""

import asyncio
import json
import logging
import urllib.parse
from dataclasses import dataclass
from typing import Optional

from aries_askar import AskarError, AskarErrorCode, Store

from ..core.error import ProfileDuplicateError, ProfileError, ProfileNotFoundError
from ..core.profile import Profile
from ..utils.env import storage_path

LOGGER = logging.getLogger(__name__)

# Error message constants
ERR_NO_STORAGE_CONFIG = "No 'storage_config' provided for postgres store"
ERR_NO_STORAGE_CREDS = "No 'storage_creds' provided for postgres store"


class AskarStoreConfig:
    """Helper for handling Askar store configuration."""

    DEFAULT_KEY = ""
    DEFAULT_KEY_DERIVATION = "kdf:argon2i:mod"
    SUPPORTED_STORAGE_TYPES = ("sqlite", "postgres")

    def __init__(self, config: Optional[dict] = None):
        """Initialize store configuration."""
        config = config or {}

        self.auto_recreate = config.get("auto_recreate", False)
        self.auto_remove = config.get("auto_remove", False)

        self.key = config.get("key", self.DEFAULT_KEY)
        self.key_derivation_method = (
            config.get("key_derivation_method") or self.DEFAULT_KEY_DERIVATION
        )
        self.rekey = config.get("rekey")
        self.rekey_derivation_method = (
            config.get("rekey_derivation_method") or self.DEFAULT_KEY_DERIVATION
        )
        self.name = config.get("name") or Profile.DEFAULT_NAME

        self.storage_config = config.get("storage_config")
        self.storage_creds = config.get("storage_creds")

        storage_type = config.get("storage_type") or "sqlite"
        if storage_type == "default":
            storage_type = "sqlite"
        elif storage_type == "postgres_storage":
            storage_type = "postgres"
        if storage_type not in self.SUPPORTED_STORAGE_TYPES:
            raise ProfileError(f"Unsupported storage type: {storage_type}")
        self.storage_type = storage_type

    def get_uri(self, create: bool = False, in_memory: Optional[bool] = False) -> str:
        """Construct the storage URI."""
        if self.storage_type == "sqlite":
            return self._build_sqlite_uri(in_memory, create)
        elif self.storage_type == "postgres":
            return self._build_postgres_uri()
        raise ProfileError(f"Unsupported storage type: {self.storage_type}")

    def _build_sqlite_uri(self, in_memory: Optional[bool], create: bool) -> str:
        if in_memory:
            return "sqlite://:memory:"
        path = storage_path("wallet", self.name, create=create).as_posix()
        return f"sqlite://{urllib.parse.quote(f'{path}/sqlite.db')}"

    def _build_postgres_uri(self) -> str:
        config, creds = self._validate_postgres_config()

        account = urllib.parse.quote(creds["account"])
        password = urllib.parse.quote(creds["password"])
        db_name = urllib.parse.quote(self.name)

        uri = f"postgres://{account}:{password}@{config['url']}/{db_name}"

        params = {}
        if "connection_timeout" in config:
            params["connect_timeout"] = config["connection_timeout"]
        if "max_connections" in config:
            params["max_connections"] = config["max_connections"]
        if "min_idle_count" in config:
            params["min_connections"] = config["min_idle_count"]
        if "admin_account" in creds:
            params["admin_account"] = creds["admin_account"]
        if "admin_password" in creds:
            params["admin_password"] = creds["admin_password"]

        if params:
            uri += "?" + urllib.parse.urlencode(params)

        return uri

    def _validate_postgres_config(self):
        if not self.storage_config:
            raise ProfileError(ERR_NO_STORAGE_CONFIG)
        if not self.storage_creds:
            raise ProfileError(ERR_NO_STORAGE_CREDS)

        try:
            config = json.loads(self.storage_config)
            creds = json.loads(self.storage_creds)
        except json.JSONDecodeError as e:
            raise ProfileError("Invalid JSON in storage config or creds") from e

        if "url" not in config:
            raise ProfileError("Missing 'url' in postgres storage_config")
        if "account" not in creds:
            raise ProfileError("Missing 'account' in postgres storage_creds")
        if "password" not in creds:
            raise ProfileError("Missing 'password' in postgres storage_creds")

        return config, creds

    async def remove_store(self):
        """Remove the store if it exists."""
        try:
            await Store.remove(self.get_uri())
        except AskarError as err:
            if err.code == AskarErrorCode.NOT_FOUND:
                raise ProfileNotFoundError(f"Store '{self.name}' not found")
            raise ProfileError("Error removing store") from err

    async def _handle_open_error(self, err: AskarError, retry=False):
        if err.code == AskarErrorCode.BACKEND:
            LOGGER.warning(
                "Askar backend error: %s. This may indicate multiple instances "
                "attempting to create the same store at the same time or a misconfigured "
                "backend.",
                err,
            )
            await asyncio.sleep(0.5)  # Wait before retrying
            return
        elif err.code == AskarErrorCode.DUPLICATE:
            raise ProfileDuplicateError(f"Duplicate store '{self.name}'")
        elif err.code == AskarErrorCode.NOT_FOUND:
            raise ProfileNotFoundError(f"Store '{self.name}' not found")
        elif retry and self.rekey:
            return
        raise ProfileError("Error opening store") from err

    async def _attempt_store_open(self, uri: str, provision: bool):
        if provision:
            return await Store.provision(
                uri,
                self.key_derivation_method,
                self.key,
                recreate=self.auto_recreate,
            )
        store = await Store.open(uri, self.key_derivation_method, self.key)
        if self.rekey:
            await Store.rekey(store, self.rekey_derivation_method, self.rekey)
        return store

    def _finalize_open(self, store, provision: bool) -> "AskarOpenStore":
        return AskarOpenStore(self, provision, store)

    async def open_store(
        self, provision: bool = False, in_memory: Optional[bool] = False
    ) -> "AskarOpenStore":
        """Open or provision the store based on configuration."""
        uri = self.get_uri(create=provision, in_memory=in_memory)

        for attempt in range(1, 4):
            LOGGER.debug("Store open attempt %d/3", attempt)
            try:
                store = await self._attempt_store_open(uri, provision)
                LOGGER.debug("Store opened successfully on attempt %d", attempt)
                return self._finalize_open(store, provision)
            except AskarError as err:
                LOGGER.debug(
                    "AskarError during store open attempt %d/3: %s", attempt, err
                )
                await self._handle_open_error(err, retry=True)

        raise ProfileError("Failed to open or provision store after retries")


@dataclass
class AskarOpenStore:
    """Handle and metadata for an opened Askar store."""

    config: AskarStoreConfig
    created: bool
    store: Store

    @property
    def name(self) -> str:
        """Accessor for the store name."""
        return self.config.name

    async def close(self):
        """Close and optionally remove the store."""
        if self.store:
            await self.store.close(remove=self.config.auto_remove)
            self.store = None



"""Module docstring."""

CATEGORY = "credential"

IDX_CRED_ON_ITEM_ID = "ON credential_record_v0_1 (item_id);"
IDX_CRED_ON_NAME = "ON credential_record_v0_1 (name);"
IDX_CRED_ON_VALUE = "ON credential_record_v0_1 (value);"

SCHEMAS = {
    "sqlite": [
        """
        CREATE TABLE IF NOT EXISTS credential_record_v0_1 (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            item_id INTEGER NOT NULL,
            name TEXT NOT NULL,
            value TEXT,
            FOREIGN KEY (item_id) REFERENCES items(id) 
                ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        "CREATE INDEX IF NOT EXISTS idx_credential_record_item_id_v0_1 "
        + IDX_CRED_ON_ITEM_ID,
        "CREATE INDEX IF NOT EXISTS idx_credential_record_item_name_v0_1 "
        + IDX_CRED_ON_NAME,
        "CREATE INDEX IF NOT EXISTS idx_credential_record_value_v0_1 "
        + IDX_CRED_ON_VALUE,
    ],
    "postgresql": [
        """
        CREATE TABLE IF NOT EXISTS credential_record_v0_1 (
            id SERIAL PRIMARY KEY,
            item_id INTEGER NOT NULL,
            name TEXT NOT NULL,
            value TEXT,
            CONSTRAINT fk_item_id FOREIGN KEY (item_id) 
            REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        "CREATE INDEX IF NOT EXISTS idx_credential_record_item_id_v0_1 "
        + IDX_CRED_ON_ITEM_ID,
        "CREATE INDEX IF NOT EXISTS idx_credential_record_item_name_v0_1 "
        + IDX_CRED_ON_NAME,
        "CREATE INDEX IF NOT EXISTS idx_credential_record_value_v0_1 "
        + IDX_CRED_ON_VALUE,
    ],
    "mssql": [
        """
        CREATE TABLE credential_record_v0_1 (
            id INT IDENTITY(1,1) PRIMARY KEY,
            item_id INT NOT NULL,
            name NVARCHAR(MAX) NOT NULL,
            value NVARCHAR(MAX),
            CONSTRAINT fk_item_id FOREIGN KEY (item_id) 
            REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        "CREATE NONCLUSTERED INDEX idx_credential_record_item_id_v0_1 "
        + IDX_CRED_ON_ITEM_ID,
        "CREATE NONCLUSTERED INDEX idx_credential_record_item_name_v0_1 "
        + IDX_CRED_ON_NAME,
        "CREATE NONCLUSTERED INDEX idx_credential_record_value_v0_1 " + IDX_CRED_ON_VALUE,
    ],
}

DROP_SCHEMAS = {
    "sqlite": [
        "DROP INDEX IF EXISTS idx_credential_record_value_v0_1;",
        "DROP INDEX IF EXISTS idx_credential_record_item_name_v0_1;",
        "DROP INDEX IF EXISTS idx_credential_record_item_id_v0_1;",
        "DROP TABLE IF EXISTS credential_record_v0_1;",
    ],
    "postgresql": [
        "DROP INDEX IF EXISTS idx_credential_record_value_v0_1;",
        "DROP INDEX IF EXISTS idx_credential_record_item_name_v0_1;",
        "DROP INDEX IF EXISTS idx_credential_record_item_id_v0_1;",
        "DROP TABLE IF EXISTS credential_record_v0_1 CASCADE;",
    ],
    "mssql": [
        "DROP INDEX IF EXISTS idx_credential_record_value_v0_1 "
        "ON credential_record_v0_1;",
        "DROP INDEX IF EXISTS idx_credential_record_item_name_v0_1 "
        "ON credential_record_v0_1;",
        "DROP INDEX IF EXISTS idx_credential_record_item_id_v0_1 "
        "ON credential_record_v0_1;",
        "DROP TABLE IF EXISTS credential_record_v0_1;",
    ],
}

COLUMNS = ["name", "value"]



import pytest

from acapy_agent.database_manager.db_types import KeyAlg, SeedMethod
from acapy_agent.database_manager.key import Key


def test_key_not_implemented_classmethods_and_ops():
    with pytest.raises(NotImplementedError):
        Key.generate(KeyAlg.A128GCM)
    with pytest.raises(NotImplementedError):
        Key.from_seed(KeyAlg.A128GCM, b"seed", method=SeedMethod.BlsKeyGen)
    with pytest.raises(NotImplementedError):
        Key.from_secret_bytes(KeyAlg.A128GCM, b"secret")
    with pytest.raises(NotImplementedError):
        Key.from_public_bytes(KeyAlg.A128GCM, b"public")
    with pytest.raises(NotImplementedError):
        Key.from_jwk({"kty": "oct"})

    k = Key(handle="h1")
    with pytest.raises(NotImplementedError):
        k.convert_key(KeyAlg.A128GCM)
    with pytest.raises(NotImplementedError):
        k.key_exchange(KeyAlg.A128GCM, Key("h2"))
    with pytest.raises(NotImplementedError):
        k.sign_message(b"msg")
    with pytest.raises(NotImplementedError):
        k.verify_signature(b"msg", b"sig")


def test_key_placeholders_and_repr():
    k = Key(handle="h1")
    assert k.handle == "h1"
    assert k.algorithm == KeyAlg.A128GCM
    assert k.ephemeral is False
    assert k.get_public_bytes() == b"public_bytes_placeholder"
    assert k.get_secret_bytes() == b"secret_bytes_placeholder"
    assert k.get_jwk_public() == "jwk_public_placeholder"
    assert k.get_jwk_secret() == b"jwk_secret_placeholder"
    assert k.get_jwk_thumbprint() == "jwk_thumbprint_placeholder"
    assert k.aead_params() == "AeadParams placeholder"
    assert k.aead_random_nonce() == b"nonce_placeholder"
    assert k.aead_encrypt(b"m") == "Encrypted placeholder"
    assert k.aead_decrypt(b"c", nonce=b"n") == b"decrypted placeholder"
    assert k.wrap_key(Key("h2")) == "Encrypted placeholder"
    r = repr(k)
    assert "Key(" in r and "handle=h1" in r
